/////////////////////////////////////////////////////////////
// Flash Plugin and Player
// Copyright (C) 1998 Olivier Debon
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
///////////////////////////////////////////////////////////////
//  Author : Olivier Debon  <odebon@club-internet.fr>
//  

#include <stdio.h>
#include "program.h"
#include "displaylist.h"
#include "graphic.h"
#include "sound.h"

#define NOTHING  0x0
#define WAKEUP   0x1
#define GOTO     0x2
#define REFRESH  0x4

static char *rcsid = "$Id: program.cc,v 1.13 1998/11/30 20:33:54 olivier Exp $";

#define PRINT 0

int debug = 0;

Program::Program(long n)
{
	long f;

	dl = new DisplayList;

	nbFrames = n;
	frames = new Frame[n];
	currentFrame = 0;
	nextFrame = currentFrame;
	getUrl = ( void (*)(char *,char *, void *))0;
	getUrlClientData = 0;
	for(f = 0; f < n; f++)
	{
		frames[f].controls = 0;
		frames[f].label = "";
	}

	currentSound = 0;
	movieStatus = MoviePlay;
	refresh = 1;
	sprite = 0;
	settings = 0;
}

Program::~Program()
{
	delete dl;
	delete frames;
}

DisplayList *
Program::getDisplayList()
{
	return dl;
}

void
Program::gotoFrame(GraphicDevice *gd, long frame)
{
	long f;

	dl->clearList();

	for(f=0; f <= frame; f++) {
		runFrame(gd, f,0);
	}
}

long
Program::runFrame(GraphicDevice *gd, long f, long action)
{
	Control		*ctrl;
	Character	*character;
	Matrix		*matrix;
	Cxform		*cxform;
	long		 status = NOTHING;
	long		 update = 0;

#if PRINT&1
	if (action) printf("Prog %x : Frame N %d\n", this, f);
#endif
	for(ctrl = frames[f].controls; ctrl; ctrl = ctrl->next)
	{
		switch (ctrl->type)
		{
			case ctrlPlaceObject:
			case ctrlPlaceObject2:
				character = 0;
				matrix = 0;
				cxform = 0;
				if (ctrl->flags & placeHasCharacter) {
					character = ctrl->character;
				}
				if (ctrl->flags & placeHasMatrix) {
					matrix = &ctrl->matrix;
				}
				if (ctrl->flags & placeHasColorXform) {
					cxform = &ctrl->cxform;
				}
				if (!ctrl->clipDepth) {	// Ignore
					dl->placeObject(character, ctrl->depth, matrix, cxform);
					update = 1;
				}
				break;
			case ctrlRemoveObject:
				character = ctrl->character;

				if (!character) break;	// Should not happen

				dl->removeObject(character, ctrl->depth);
				if (action) {
					if (character->hasEventHandler()) {
						gd->clearHitTest(character->getTagId());
					}
					character->reset();
					update = 1;
				}
				break;
			case ctrlRemoveObject2:
				character = dl->removeObject(ctrl->depth);
				if (character && action) {
					if (character->hasEventHandler()) {
						gd->clearHitTest(character->getTagId());
					}
					character->reset();
					update = 1;
				}
				break;
		// Actions
			case ctrlDoAction:
				if (action) {
					status = doAction(ctrl->actionRecords);
				}
				break;
			case ctrlStartSound:
				if (action) {
					if (currentSound) {
						currentSound->stop();
					}
					currentSound = (Sound *)ctrl->character;
				}
				break;
			case ctrlStopSound:
				if (action && currentSound) {
					currentSound->stop();
				}
				break;
			case ctrlBackgroundColor:
				dl->setBackgroundColor(&ctrl->color);
				if (action) {
					gd->setBackgroundColor(ctrl->color);
				}
				break;
		}
	}
	if (status & GOTO) {
		gotoFrame(gd, nextFrame);
		update = 1;
	}

#if PRINT&1
	if (action) printf("Frame N %d ready\n", f);
#endif
	return update;
}

long
Program::nestedMovie(GraphicDevice *gd, Matrix *mat)
{
	if (movieStatus == MoviePlay) {
		// Movie Beeing Played
		advanceFrame();
		if (currentFrame == 0) {
			dl->clearList();
		}
		runFrame(gd, currentFrame);
		if (nbFrames == 1) {
			pauseMovie();
		}
	}

	sprite = dl->render(gd,mat);

	return (sprite || movieStatus == MoviePlay);
}

long
Program::processMovie(GraphicDevice *gd)
{
	if (currentSound) {
#if PRINT&1
		printf("Sound\n");
#endif
		if (!currentSound->play()) {
			currentSound = 0;
		}
	}

#if PRINT&1
	printf("Prog %x : Current = %d     Next = %d\n", this, currentFrame, nextFrame);
#endif

	if (movieStatus == MoviePlay) {
		// Movie Beeing Played
		advanceFrame();
		if (currentFrame == 0) {
			gd->resetHitTest();
			dl->clearList();
		}
		refresh |= runFrame(gd, currentFrame);
		if (nextFrame == nbFrames && ((settings & PLAYER_LOOP) == 0)) {
			pauseMovie();
		}
	}

	if (sprite || refresh) {
		gd->clearCanvas();
		sprite = dl->render(gd);
		refresh = 0;
		gd->displayCanvas();
	}

	XFlush(gd->dpy);
	return (sprite || movieStatus == MoviePlay || currentSound);
}

long
Program::handleEvent(GraphicDevice *gd, XEvent *event)
{
	ActionRecord	*action;
	long		 status = NOTHING;

	if (event) {
		// Exposure event should be passed to objects, but not for the moment
		if (event->type == Expose && event->xexpose.count == 0) {
			gd->displayCanvas();
		} else if (event->type != NoExpose) {
			action = dl->processEvent(gd,event);
			status = doAction(action);
			if (status & REFRESH) {
				status |= WAKEUP;
				refresh = 1;
			}
			if (status & GOTO) {
				gd->resetHitTest();
				gotoFrame(gd, nextFrame);
				refresh = 1;
			}
		}
	}
	if (status) return processMovie(gd);
	return 0;
}

long
Program::doAction(ActionRecord *action)
{
	long status = NOTHING;

	while(action)
	{
		switch (action->action)
		{
			case ActionPlaySound:
#if PRINT&2
				printf("Prog %x : PlaySound\n", this);
#endif
				currentSound = action->sound;
				status |= WAKEUP;
				break;
			case ActionRefresh:
#if PRINT&2
				printf("Prog %x : Refresh\n", this);
#endif
				status |= REFRESH;
				break;
			case ActionGotoFrame:
#if PRINT&2
				printf("Prog %x : GotoFrame %d\n", this, action->frameIndex);
#endif
				nextFrame = action->frameIndex;
				status |= WAKEUP|GOTO;
				break;
			case ActionGetURL:
#if PRINT&2
				printf("Prog %x : GetURL %s target = %s\n", this, action->url, action->target);
#endif
				if (getUrl) {
					getUrl(action->url, action->target, getUrlClientData);
				}
				break;
			case ActionNextFrame:
				nextFrame = currentFrame+1;
				status |= WAKEUP;
				break;
			case ActionPrevFrame:
				nextFrame = currentFrame-1;
				status |= WAKEUP|GOTO;
				break;
			case ActionPlay:
#if PRINT&2
				printf("Prog %x : Play\n", this);
#endif
				movieStatus = MoviePlay;
				if (currentFrame == nextFrame) advanceFrame();
				status |= WAKEUP;
				break;
			case ActionStop:
#if PRINT&2
				printf("Prog %x : Stop\n", this);
#endif
				movieStatus = MoviePaused;
				nextFrame = currentFrame;
				break;
			case ActionToggleQuality:
				break;
			case ActionStopSounds:
				if (currentSound) {
					currentSound->stop();
				}
				break;
			case ActionWaitForFrame:
				break;
			case ActionSetTarget:
				break;
			case ActionGoToLabel:
#if PRINT&2
				printf("Prog %x : GotoFrame '%s'\n", this, action->frameLabel);
#endif
				nextFrame = searchFrame(action->frameLabel);
				status |= WAKEUP|GOTO;
				break;
		}
		action = action->next;
	}
	return status;
}

void
Program::setCurrentFrameLabel(char *label)
{
	frames[currentFrame].label = label;
}

void
Program::rewindMovie()
{
	currentFrame = 0;
	nextFrame = 0;
}

void
Program::pauseMovie()
{
	movieStatus = MoviePaused;
	nextFrame = currentFrame;
}

void
Program::continueMovie()
{
	movieStatus = MoviePlay;
}

void
Program::nextStepMovie()
{
	if (movieStatus == MoviePaused) {
		advanceFrame();
	}
}

void
Program::advanceFrame()
{
	currentFrame = nextFrame;
	nextFrame = currentFrame+1;
	if (currentFrame == nbFrames) {
		currentFrame = 0;
		nextFrame = 1;
	}
}

void Program::setCurrentFrame(long n)
{
	currentFrame = n;
}

long Program::getFrame()
{
	return currentFrame;
}

void
Program::addControlInCurrentFrame(Control *ctrl)
{
	Control *c;

	ctrl->next = 0;
	if (frames[currentFrame].controls == 0) {
		frames[currentFrame].controls = ctrl;
	} else {
		for(c = frames[currentFrame].controls; c->next; c = c->next);
		c->next = ctrl;
	}
}

void
Program::setGetUrlMethod( void (*func)(char *, char *, void *), void *clientData)
{
	getUrl = func;
	getUrlClientData = clientData;
}

void
Program::modifySettings(long flags)
{
	settings = flags;
}

long
Program::searchFrame(char *label)
{
	long f;

	for(f=0; f < nbFrames; f++)
	{
		if (frames[f].label && !strcmp(label,frames[f].label)) {
			return f;
		}
	}
	return 0;
}
